---
title: Binary search
nav_order: 1
---

Let's re-analyze the runtime for binary search on a sorted array, but this time looking at the recursive implementation. The `binarySearch` method below returns the index of the search `target` or -1 if the `target` is not in the `sorted` array.

<div style="--aspect-ratio: 720 / 405; --add-height: 29px"><iframe src="https://docs.google.com/presentation/d/e/2PACX-1vTir_A7eOXUQ5UhLEN9aG4bX466Xgh_ViaZACF_UlpcJZaWfsWIJCqu4b-KpCytEAJT7vpQBeUpvRFb/embed?start=false&loop=false&delayms=3000" frameborder="0" width="720" height="434" allowfullscreen="true" mozallowfullscreen="true" webkitallowfullscreen="true"></iframe></div>
```java
static int binarySearch(int[] sorted, int target) {
    return binarySearch(sorted, target, 0, sorted.length);
}

static int binarySearch(int[] sorted, int target, int low, int high) {
    if (low > high) {
        return -1;
    }
    int mid = low + (high - low) / 2;
    if (sorted[mid] < target) {
        return binarySearch(sorted, target, mid + 1, high);
    } else if (sorted[mid] > target) {
        return binarySearch(sorted, target, low, mid - 1);
    } else {
        return mid;
    }
}
```

<details markdown="1">
<summary>What is the best-case order of growth of the runtime?</summary>

In the best case, the `target` is the exact middle item in the `sorted` array regardless of the length of the array. Since it's not necessary to recurse to either the left or right, the runtime is just constant.
</details>

It turns out that the worst-case order of growth of the runtime for `binarySearch` is logarithmic with respect to the length of the `sorted` array. In each recursive call of the loop, half of the remaining elements under consideration can be ignored. In other words, the number of recursive calls is given by the answer to the question, "How many times do we need to divide *N* (length of the `sorted` array) by 2 until only 1 element remains?" This is the definition of the **binary logarithm** (log<sub>2</sub> or simply lg).

Logarithm base
: Computer scientists often don't specify the logarithm base. A common assumption is to work in base-2 (binary logarithm) since it corresponds to how bits are represented in a computer. For the purpose of asymptotic analysis, the log base doesn't affect the result because all log bases are within a [constant multiplicative factor of each other](https://en.wikipedia.org/wiki/List_of_logarithmic_identities#Changing_the_base).

How might we prove that the worst-case order of growth of the runtime of `binarySearch` is in *Θ*(log *N*)? We have a formal, mathematical definition for big-theta notation, so it's nice to be able to use it!

**Recurrence relations** are recursive functions that model non-recursive work and recursive work.

Recurrence relation for the worst-case runtime of `binarySearch`
: *T*(*N*) = *T*(*N*/2) + *c* for *N* > 1
: *T*(1) = *d*

*c* represents the constant time spent on non-recursive work, such as comparing `low < high`, computing `mid`, and comparing the `target` with `sorted[mid]`. *T*(1) = *d* represents the base case, which takes a different amount of constant time to compare `low < high` and immediately `return -1`. The time spent on recursive work is modeled by *T*(*N*/2) because a recursive call to `binarySearch` will examine either the lower half or upper half of the remaining *N* items.

Our goal is to rewrite this recurrence relation in a [closed-form expression](https://en.wikipedia.org/wiki/Closed-form_expression) that's compatible with asymptotic notation definitions. One approach is to **unroll the recurrence**: plug the recurrence back into itself until the recursion is done.

- *T*(*N*) = *T*(*N*/2) + *c*
- *T*(*N*) = *T*(*N*/4) + *c* + *c*
- *T*(*N*) = *T*(*N*/8) + *c*(3)
- …
- *T*(*N*) = *T*(*N*/*N*) + *c*(log *N*)
- *T*(*N*) = *d* + *c*(log *N*)

We can then apply big-theta notation to describe the order of growth of the closed-form expression for *T*(*N*), the worst-case runtime of `binarySearch`.
